from eubi_bridge import ebridge
from eubi_bridge import ebridge_base
from eubi_bridge import fileset_io
from eubi_bridge.base import scale, writers
from eubi_bridge.ngff import defaults, multiscales
from eubi_bridge.utils import convenience, dask_client_plugins
from eubi_bridge.utils.logging_config import setup_logging, get_logger
import logging

# Set up default logging configuration when the package is imported
setup_logging(
    log_level=logging.INFO,
    log_file=None,  # Set to a path to enable file logging
    console=True
)

# Create a logger for the package
logger = get_logger(__name__)
logger.debug("eubi_bridge package initialized")

__all__ = [
    'ebridge',
    'ebridge_base',
    'fileset_io',
    'scale',
    'writers',
    'defaults',
    'multiscales',
    'convenience',
    'dask_client_plugins',
    'setup_logging',
    'get_logger',
]
